/*

*/

import server.*;

import java.time.format.DateTimeFormatter;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;

import java.util.Properties;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.IOException;

/**
   This is an example of how you can use the HttpApplicationServer
   framework in the server package to implement a very simple
   web application.

   The web application implemented here is an "addition server".
   Given a URL like
      http://localhost:8080/?x=61&y=62
   the web application will add the two given parameters and
   return 123.
*/
public class SampleServer_ver2 extends HttpMethodsAdaptor
{
   /**
      Give this sample server an implementation of GET.

      This GET method expects two URL query parameters
      that are integers and returns the sum of those
      two numbers.
   */
   @Override
   public void doGet(HttpRequest req, HttpResponse res)
   {
      System.out.println("SERVER: GET addition method:");

      if (! req.resourceName.equals("/") )
      {  // https://httpwg.org/specs/rfc9110.html#status.204
         sendErrorResponse(req, res,
                "204",
                "No Content",
                "");
         return;
      }

      // Get the URL's query string from the req object.
      final String query = req.queryString;
      // Parse it.
      final int i1 = query.indexOf("=");
      final int i2 = query.indexOf("&");
      final int i3 = query.lastIndexOf("=");
      String p1 = query.substring(i1 + 1, i2);
      String p2 = query.substring(i3 + 1);
      int x = Integer.parseInt(p1);
      int y = Integer.parseInt(p2);

      // Build the response body.
      String body = "<!doctype html>\r\n" +
                    "<html>\r\n" +
                    "<head>\r\n" +
                    "<title>Addition Server</title>\r\n" +
                    "</head>\r\n" +
                    "<body>\r\n" +
                    "<h1>Your addition results.</h1>\r\n" +
                    "<p>" + p1 + " + " + p2 + " = " + (x+y) + "</p>\r\n" +
                    "<hr>\r\n" +
                    "</body>\r\n" +
                    "</html>\r\n";

      // Add the response line.
      res.setResponseLine("HTTP/1.1 200 OK");

      // Add the response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      res.addHeader("Date: " + date);
      res.addHeader("Server: SampleServer_ver2.java");
      res.addHeader("Connection: close");
      res.addHeader("Content-type: text/html");
      res.addHeader("Content-Length: " + body.length());

      // Add the entity body.
      res.setEntityBody( body.getBytes() );

      return;
   }



   /**
      The main() method configures the server using
      the httpserver.properties file and command-line
      arguments.
   */
   public static void main(String[] args)
   {
      String serverRoot = "public_html"; // default value
      int portNumber = 8080;             // default value

      final Properties properties = new Properties();
      try( FileInputStream fis =
             new FileInputStream(
                new File("httpserver.properties")) )
      {
         properties.load( fis );
         serverRoot = properties.getProperty("root");
         final String port = properties.getProperty("port");
         if (port != null) portNumber = Integer.parseInt(port);
      }
      catch (FileNotFoundException e)
      {
         /* ignore FileNotFoundException */
      }
      catch (IOException e)
      {
         e.printStackTrace(System.out);
         System.exit(-1);
      }

      if (0 < args.length)
      {
         portNumber = Integer.parseInt(args[0]);
      }
      if (1 < args.length)
      {
         serverRoot = args[1];
      }

      // Instantiate and start up an instance of the server.
      new HttpApplicationServer(portNumber,
                                serverRoot,
                                new SampleServer_ver2()); // HTTP methods
   }
}
