/*

*/

package server;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.net.*;
import java.time.format.DateTimeFormatter;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;

/**
   This class turns the HttpApplicationServer framework
   into a static HTTP server that supports most of the
   HTTP methods (GET, HEAD, POST, PUT, OPTIONS).
*/
public class HttpServerMethods extends HttpMethodsAdaptor
{
   /**
      Dispatch static and dynamic GET requests.
   */
   @Override
   public void doGet(HttpRequest req, HttpResponse res)
   {
      // Make sure the resource exits and is readable.
      final File file = new File(server.serverRoot + req.resourceName);
      if ( ! file.exists() || ! file.canRead() )
      {
         if (! file.exists())
         {
            sendErrorResponse(req, res,
                "404", "Not Found",
                "The requested URL "+req.resourceName+" was not found on this server.");
         }
         else
         {
            sendErrorResponse(req, res,
                "403", "Forbidden",
                "The requested URL "+req.resourceName+" is not readable.");
         }

         return;  // Finished with this request.
      }

      if ( file.isDirectory() )
      {
         // Check for existence of an index.html file.
         final File file2 = new File(server.serverRoot + req.resourceName + "index.html");
         if ( file2.exists() && file2.canRead() )
         {
            req.resourceName = req.resourceName + "index.html";
            doStaticGet(req, res, file2);
         }
         else
         {
            doDynDirListing(req, res, file);
         }
      }
      else if ( req.resourceName.endsWith(".class")
             || req.resourceName.endsWith(".exe") )
      {
         doCGIGet(req, res, file);
      }
      else // the file exists, so send it to the client
      {
         doStaticGet(req, res, file);
      }
   }



   /**
      Send a static resource to the client.
   */
   public void doStaticGet(HttpRequest req,
                           HttpResponse res,
                           File file)
   {
      System.out.println("SERVER: doStaticGet: ");

      final int numOfBytes = (int)file.length();
      try
      {
         final FileInputStream inFile = new FileInputStream(file);
         final byte[] fileInBytes = new byte[numOfBytes];
         inFile.read(fileInBytes);

         // Add the response line.
         res.setResponseLine("HTTP/1.1 200 OK");
         // Add the response headers.
         final String date = OffsetDateTime.now(ZoneOffset.UTC)
                               .format(DateTimeFormatter.RFC_1123_DATE_TIME);
         res.addHeader("Date: " + date);
         res.addHeader("Server: HttpApplicationServer.java");
         res.addHeader("Connection: close");
         final String contentType;
         if ( req.resourceName.endsWith(".html") || req.resourceName.endsWith(".htm") )
            contentType = "text/html";
         else if ( req.resourceName.endsWith(".css") )
            contentType = "text/css";
         else if ( req.resourceName.endsWith(".js") )
            contentType = "application/javascript";
         else if ( req.resourceName.endsWith(".jpg") || req.resourceName.endsWith(".jpeg") )
             contentType = "image/jpeg";
         else if ( req.resourceName.endsWith(".gif") )
            contentType = "image/gif";
         else if ( req.resourceName.endsWith(".png") )
             contentType = "image/png";
         else  // default content type
            contentType = "text/plain";
         res.addHeader("Content-Type: " + contentType);
         res.addHeader("Content-Length: " + numOfBytes);

         // Add the entity body.
         if ( ! req.requestMethod.equals("HEAD") )
         {
            res.entityBody = fileInBytes;
         }
      }
      catch (FileNotFoundException e)
      {
         System.out.println("SERVER: doStaticGet: Unable to find file.");
         System.out.println( e );
      }
      catch (IOException e)
      {
         System.out.println("SERVER: doStaticGet: Unable to read file.");
         System.out.println( e );
      }
   }


   /**
      This method should implement the CGI GET protocol for
      inter-process communication between the server process,
      its child process, and the client process.
   */
   public void doCGIGet(HttpRequest req,
                        HttpResponse res,
                        File file)
   {
      // This is just a sketch of an implementation of the cgi GET method.
      //
      // This method should run the resource program named resourceName
      // and pass it the querryString data. The resource program would
      // then complete the transaction with the client.
      //
      // Instead, this sketch just logs the querryString to the server's
      // stdout and then passes the rest of the request to doStaticGet.
      // From the client's perspective, a cgi GET is the same as a
      // static GET.

      System.out.println("SERVER: doCGIGet: ");

      // Log the query string to stdout.
      System.out.println( "===> Received a CGI GET request with the following query string.\n" );
      System.out.println( "===> " + req.queryString + "\n" );
      System.out.println( "===> Processing as a normal GET request.\n" );

      // Send the client the static resourse.
      doStaticGet(req, res, file);
   }


   /**
      This method should be implemented to send the client a
      html page that is a listing of the given directory.
   */
   public void doDynDirListing(HttpRequest req,
                               HttpResponse res,
                               File file)
   {
      sendErrorResponse(req, res,
          "400", "Bad Request",
          "The requested URL, "+req.resourceName+", does not contain a file name.");
   }


   /**
      Dispatch the different kinds of POST requests.
   */
   @Override
   public void doPost(HttpRequest req, HttpResponse res)
   {
      // Make sure the resource exits and is readable.
      final File file = new File(server.serverRoot + req.resourceName);
      if ( ! file.exists() || ! file.canRead() )
      {
         if (! file.exists())
         {
            sendErrorResponse(req, res,
                "404", "Not Found",
                "The requested URL "+req.resourceName+" was not found on this server.");
         }
         else
         {
            sendErrorResponse(req, res,
                "403", "Forbidden",
                "The requested URL "+req.resourceName+" is not readable.");
         }

         return;  // Finished with this request.
      }

      if ( req.resourceName.endsWith(".class")
        || req.resourceName.endsWith(".exe") )
      {
         doCGIPost(req, res, file);
      }
      else if ( req.resourceName.equals("/post_data") )
      {
         doBuiltInPost(req, res, file);
      }
      else
      {
         doStaticPost(req, res, file);
      }
   }


   /**
      This method should implement the CGI POST protocol for
      inter-process communication between the server process,
      its child process, and the client.
   */
   public void doCGIPost(HttpRequest req,
                         HttpResponse res,
                         File file)
   {
      System.out.println("SERVER: doCGIPost: ");

      // Log the entity body to stdout.
      final String entityBody = new String(req.entityBody);
      System.out.println( "===> " + entityBody );

      // We need to launch the Java program called resourceName.
      // Get the path of the java.exe program that started this server's JVM.
      String javaPath = ProcessHandle.current()
                        .info()
                        .command()
                        .orElse(null);
      if(javaPath == null)
      {
         System.out.println("SERVER: doCGIPost: Could not get the path to java.exe.");
         return;
      }
      System.out.println("SERVER: doCGIPost: Using this JVM for POST: " + javaPath);

      // Configure the ProcessBuilder.
      try
      {
         final String resourceClass = req.resourceName
                                     .substring(0, req.resourceName.lastIndexOf("."));
         final String classFile = new File(server.serverRoot + resourceClass)
                                      .getCanonicalFile()
                                      .getAbsolutePath();
         System.out.println("SERVER: doCGIPost: Running this class file: " + classFile);

         final ProcessBuilder pb = new ProcessBuilder(
                                         javaPath,
                                         "-cp",
                                         new File(server.serverRoot).getCanonicalFile().getAbsolutePath(),
                                         resourceClass.substring(1));
         System.out.println( pb.command() );

         // Start a new java process.
         final Process p = pb.start();
         System.out.println("SERVER: doCGIPost: Launced this child process: " + p);
         // Write POST data to the input stream of the process.
         try (PrintStream out = new PrintStream(
                                   p.getOutputStream()))
         {
            out.print( new String(req.entityBody) );
            out.flush();
         }

         // Add the response line.
         res.setResponseLine("HTTP/1.1 200 OK");

         // Add the first response headers.
         final String date = OffsetDateTime.now(ZoneOffset.UTC)
                               .format(DateTimeFormatter.RFC_1123_DATE_TIME);
         res.addHeader("Date: " + date);
         res.addHeader("Server: HttpApplicationServer.java");
         res.addHeader("Connection: close");

         // Read the standard output stream of the process
         // and store the results as the entity body.
         System.out.println("====> doCGIPost reading response from child process.");
         String responseBody = "";
         try (BufferedReader in = new BufferedReader(
                                    new InputStreamReader(
                                      p.getInputStream())))
         {
            String oneLine;

            while ((oneLine = in.readLine()) != null && !oneLine.isEmpty() )
            {
               res.addHeader(oneLine);
               System.out.println("<" + oneLine);
            }
            while ((oneLine = in.readLine()) != null)
            {
               responseBody += oneLine + "\n";  // Should use StringBuilder.
               System.out.println("<" + oneLine);
            }
         }
         res.setEntityBody( responseBody.getBytes() );
         System.out.println("====> doCGIPost done reading response from child process.");

         // Read the standard error stream of the process.
         System.out.println("====> doCGIPost reading error stream from child process.");
         try (BufferedReader err = new BufferedReader(
                                     new InputStreamReader(
                                       p.getErrorStream())))
         {
            String oneLine;
            while ((oneLine = err.readLine()) != null)
            {
               System.out.println("err: " + oneLine);
            }
         }
         System.out.println("====> doCGIPost done reading errors from child process.");
      }
      catch (IOException e)
      {
         System.out.println("SERVER: doCGIPost: Error while runing " + req.resourceName);
         System.out.println( e );
      }
   }


   /**
      This method implements a simple built-in POST protocol.
      This method prints the entity body to the server's stdout
      and then sends back to the client a dynamically generated
      "thank you" web page which includes a copy of the client's
      POST data.

      In general, a method like this can do any amount of
      computation using the client's POST data as the input
      to the calculations. And the returned web page can
      then depend on those calculations.
   */
   public void doBuiltInPost(HttpRequest req,
                             HttpResponse res,
                             File file)
   {
      System.out.println("SERVER: doBuiltInPost: ");

      // Log the entity body to stdout.
      final String entityBody = new String(req.entityBody);
      System.out.println( "===> " + entityBody );

      // Build the HTTP response body.
      final String body = "<!doctype html>\r\n" +
                          "<html>\r\n" +
                          "<head>\r\n" +
                          "<title>Thank you!</title>\r\n" +
                          "</head>\r\n" +
                          "<body>\r\n" +
                          "<h1>Thank you!</h1>\r\n" +
                          "<p>Thank you for your POST data!</p>\r\n" +
                          "<p></p>\r\n" +
                          "<p>" + entityBody + "</p>\r\n" +
                          "<p></p>\r\n" +
                          "<p>We will protect it forever.</p>\r\n" +
                          "</body>\r\n" +
                          "</html>\r\n";

      // Add the response line.
      res.setResponseLine("HTTP/1.1 200 OK");

      // Add the response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      res.addHeader("Date: " + date);
      res.addHeader("Server: HttpApplicationServer.java");
      res.addHeader("Connection: close");
      res.addHeader("Content-Type: text/html");
      res.addHeader("Content-Length: " + body.length());

      // Add the entity body.
      res.setEntityBody(body.getBytes());
   }


   /**
      This method should implement a static POST protocol.
   */
   public void doStaticPost(HttpRequest req,
                            HttpResponse res,
                            File file)
   {
      // This is just a sketch of an implementation of the POST method.
      // This sketch prints the entity body to the server's stdout and
      // then sends back to the client the static resource resourceName.

      System.out.println("SERVER: doStaticPost: ");

      // Log the entity body to stdout.
      final String data = new String(req.entityBody);
      System.out.println( "===> " + data );

      // Note: We could substitute the entity data string
      //       for the queryString and then call doCGIGet().

      // Send the client the static resourse.
      doStaticGet(req, res, file);
   }


   /**
      This method should implement the HTTP OPTIONS command.
   */
   @Override
   public void doOptions(HttpRequest req, HttpResponse res)
   {
      // Add the response line.
      res.setResponseLine("HTTP/1.1 200 OK");

      // Add the response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      res.addHeader("Date: " + date);
      res.addHeader("Server: HttpApplicationServer.java");
      res.addHeader("Allow: GET,HEAD,POST,OPTIONS");
      res.addHeader("Connection: close");
      res.addHeader("Content-type: text/html");
      res.addHeader("Content-Length: 0");
   }


   /**
      This method should implement the HTTP PUT command.
   */
   @Override
   public void doPut(HttpRequest req, HttpResponse res)
   {
      // Store the entity body in the server's file system.
      final File file = new File(server.serverRoot + "/files/" + req.resourceName);

      System.out.println("SERVER: doPut: about to write file " +  file);

      PrintStream fileOut = null;
      try
      {
         fileOut = new PrintStream(file);
      }
      catch (FileNotFoundException e)
      {
            System.out.println("SERVER: doPut: Unable to find destination file.");
            System.out.println( e );
      }
      try
      {
         fileOut.write(req.entityBody);
         fileOut.flush();
         fileOut.close();
         System.out.println("SERVER: doPut: done writing file " +  file);
      }
      catch (IOException e)
      {
            System.out.println("SERVER: doPut: Unable to write destination file.");
            System.out.println( e );
      }

      // Add the response line.
      res.setResponseLine("HTTP/1.1 201 Created");

      // Add the response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      res.addHeader("Date: " + date);
      res.addHeader("Server: HttpApplicationServer.java");
      res.addHeader("Connection: close");
      res.addHeader("Content-type: text/html");
      res.addHeader("Content-Length: 0");
   }
}
